/*

Font editor for display LCD unit 
Copyleft 1997 Andrew J. Kroll (ag784@freenet.buffalo.edu)

This lets you see what your characters will look like before you code them.
It's not only a tool, but can be a fun toy! :-)

Some Parts Copyright 1991-1995 Stephen S. Richardson (prefect@wpi.edu)
based off of several versions for several hardware and software platforms
from 1991-1995.

(credit is due to Arthur C. Smith, whose joystick code I studied to learn how
to write linux device drivers, and to Linus Torvalds, for being the cool kind
of person to start the whole Linux project.. yar.)

You may modify, copy, fix, change, or delete this code, as long as my
copyright remains intact.  Well, if you delete it, it doesn't have to.  I also
ask that you send me a piece of email if you make any interesting changes.
You don't have to, but I want to see if anyone does anything cool with it.

There are no warranties, expressed or implied.  If your computer blows up, or
you blow up your sibling trying to use this, I'm not responsible.  If it works,
and all your friends thing you're cool, you must tell them it is due to this
project. ;)


*/

#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <utmp.h>
#include <math.h>
#include "release.h"
#include "lcd.h"
#define POSIX
#include <termios.h>
#define TERMIO termios

struct TERMIO tty_old, tty_new;

void lcd_init (int fd)
{
  int status;
  struct LCD_DATA_TYPE lcd;

  status=ioctl (fd, LCD_SETCONTROL);
  if (status == -1) {
    perror ("lcd");
    exit (1);
  }

  strcpy (lcd.text, LCD_INITSTRING);
  status = write (fd, &lcd, LCD_DATASIZE);
  if (status != LCD_DATASIZE) {
    perror ("lcd");
    exit (1);
  }
  usleep (1000);
}

void lcd_open (int *fd)
{
  *fd = open ("/dev/lcd0", O_WRONLY);
  if (*fd < 0) {
	perror ("lcd");
	exit (1);
      }
  lcd_init (*fd);
}



void lcd_clear (int fd)
{
  int status;
  struct LCD_DATA_TYPE lcd;

  status=ioctl (fd, LCD_SETCONTROL);
  if (status == -1) {
    perror ("lcd");
    exit (1);
  }

  strcpy (lcd.text, LCD_CLEARSTRING);
  status = write (fd, &lcd, LCD_DATASIZE);
  if (status != LCD_DATASIZE) {
    perror ("lcd");
    exit (1);
  }
}

void lcd_pos (int fd, int pos)
{
  int status;
  struct LCD_DATA_TYPE lcd;

  status=ioctl (fd, LCD_SETCONTROL);
  if (status == -1) {
    perror ("lcd");
    exit (1);
  }

  if ((pos<0)||(pos>0x40+40)) {
    printf ("lcd: position out of range\n");
    exit (1);
  }

  lcd.text[0]=128+pos;
  lcd.text[1]=0;

  status = write (fd, &lcd, LCD_DATASIZE);
  if (status != LCD_DATASIZE) {
    perror ("lcd");
    exit (1);
  }
}
 
void lcd_write (int fd, char *str)
{
  struct LCD_DATA_TYPE lcd;
  int status;

  strcpy (lcd.text, str);
  ioctl (fd, LCD_SETDATA);
  status = write (fd, &lcd, LCD_DATASIZE);
  if (status != LCD_DATASIZE) {
    perror ("lcd");
    exit (1);
  }
}

void lcd_gRAM (int fd, int add)
{
  int status;
  char datm[2]; 
  struct LCD_DATA_TYPE lcd;

  datm[1]=0;
  datm[0]=add+0x40;
  status=ioctl (fd, LCD_SETCONTROL);
  if (status == -1) {
    perror ("lcd");
    exit (1);
  }
  
  strcpy (lcd.text, datm);
  status = write (fd, &lcd, LCD_DATASIZE);
  if (status != LCD_DATASIZE) {
    perror ("lcd");
    exit (1);
  }
}

void lcd_cursor (int fd, int toggle)
{
  int status;
  char datm[2]; 
  struct LCD_DATA_TYPE lcd;

  datm[1]=0;
  datm[0]=12; /* no cursor */
  if (toggle == 1) datm[0]=15;
  status=ioctl (fd, LCD_SETCONTROL);
  if (status == -1) {
    perror ("lcd");
    exit (1);
  }
  
  strcpy (lcd.text, datm);
  status = write (fd, &lcd, LCD_DATASIZE);
  if (status != LCD_DATASIZE) {
    perror ("lcd");
    exit (1);
  }
}

void get_data (int fd, int c)
{
  	char text[81],bfr[6],key; /* display entry */
	int dta, gpos;
/*                   43210 <--- bits */
	for(dta=0;dta<8;dta++) {
		sprintf(bfr,"00000");
		for(gpos=0;gpos<5;) {
			lcd_pos(fd,9+0x40);
			sprintf(text,"%5.5s %1.1d",bfr,(int) dta);
			lcd_write (fd, text);
			lcd_pos(fd,9+0x40+gpos);
			for(key=0;key == 0;key=getchar());
			if (key == '0') {
				bfr[gpos]='0';
				gpos++;
			}
			if (key == '1') {
				bfr[gpos]='1';
				gpos++;
			}
				
		}
	/* update char gen RAM */	
	key=0;
	for(gpos=0;gpos<5;gpos++) { 
		key = (key << 1) | ((int) bfr[gpos] - (int) '0');
		/* printf("%d\n\r",key); */
	}
	lcd_gRAM(fd,dta+(c*8));
	sprintf(text,"%c",(char) key);
	lcd_write (fd, text);
	}
/* blank entry area */
	lcd_pos(fd,9+0x40);
	sprintf(text,"       ");
	lcd_write (fd, text);

}

void showthem (int fd)
{
  	char text[81], l1[41], l2[41];
  	int c=0;
  	lcd_init (fd);
  	/*            0123456789012345678901234567890123456789 */
  	/*            0123456789012345	*/
  	sprintf (l1, "                <-- not useable 16x2 -->"); 
  	sprintf (l2, "01234567        <-- not useable 16x2 -->");
  	sprintf (text, "%-40.40s%-40.40s", l1, l2);
  	lcd_write (fd, text);
	for (c=0;c<8;c++) {
    		sprintf (text, "%c", (char) c);
    		lcd_pos (fd, c);
    		lcd_write (fd, text);
    	}
	sprintf (text, "<-Cusset");
	lcd_write (fd, text);
}

void editthem (int fd)
{
	char key=0; /* keyboard crap */
  	int c=0; /* current one */
	lcd_cursor(fd,1); /* turn on cursor */
	for(;;) {
		lcd_pos(fd,c+0x40); /* position cursor to blink 
					under current char */
		key=getchar();
		if (key == 'q') break;
		if (key == 'n') c++;
		if (key == 'p') c--;
		if (c<0) c=7;
		if (c>7) c=0;
		if (key == 'e') get_data(fd, c);
	}
}

void init_tty() {
  
  bcopy(&tty_old,&tty_new,sizeof(tty_new));
  tty_new.c_iflag |= IGNBRK;
  tty_new.c_lflag=0;
  #ifdef POSIX
     tty_new.c_cflag &= ~(PARENB|CSIZE);         /* Disable parity */
     tty_new.c_cflag |= (CREAD|CS8);     /* Set character size = 8 */
  #else
     tty_new.c_line=0;
  #endif
  tty_new.c_cc[VMIN]=1;
  (unsigned) tty_new.c_cc[VTIME]=1000;
  #ifdef POSIX
     tcsetattr(0, TCSANOW, &tty_new);
  #else
     ioctl(0,TCSETA,&tty_new);
  #endif
}
                                                     


int main (void)
{
	int fd;
	
	#ifdef POSIX
	   tcgetattr(0, &tty_old);
	#else
	   ioctl(0,TCGETA,&tty_old);
	#endif
	init_tty();
	lcd_open (&fd);
	lcd_clear (fd);
  	
  	printf("\r\nSee display for current custom character"
  	" and construction of each.\r\nThis is a preview of what I have"
  	"planned for the next version... :-)\r\n\r\n"
	"All info is shown on your LCD panel\r\n\r\n"
	"keys:\r\n"
	"n next\r\n"
	"p previous\r\n"
	"e enter data (hit 0 and 1 to fill each row)\r\n"
	"q quits\r\n"
	);
	
	showthem (fd);
	editthem (fd);
	#ifdef POSIX
	   tcsetattr(0, TCSANOW, &tty_old);
	#else
	   ioctl(0,TCSETA,&tty_old);
	#endif
	printf("Done.\r\n");
	lcd_cursor(fd,0); /* turn off cursor */
	exit (0);
}
