/*

Copyleft 1997 Andrew J. Kroll (ag784@freenet.buffalo.edu)

This lets you see what your characters will look like before you code them.
It's not only a tool, but can be a fun toy! :-)

Some Parts Copyright 1991-1995 Stephen S. Richardson (prefect@wpi.edu)
based off of several versions for several hardware and software platforms
from 1991-1995.

10/23/94	S.Richardson	0.1	Initial version
10/26/94	S.Richardson	0.2	More revisions
01/??/95	S.Richardson	0.3	More revisions again.
02/02/95	S.Richardson	0.042	"release"
07/30/95	S.Richardson	0.50	"release 2" - cleanup for newer kernels.
02/20/97        Andrew J. Kroll 0.51    Real-time display and support for 16x2.
02/20/97	Andrew J. Kroll	0.51	Fixed nasty bug locks up anything on 
					IRQ "shared" with LPT card :-O
06/01/97	Andrew J. Kroll 0.51c   r/w to remove cursor problems
*/

#define DRIVER

#include "lcd.h"

static struct LCD_DATA_SAVE_TYPE LCD_DATA[LCD_MAX];	/* misc data */
static int LCD_WRITE_SEMAPHORE;	/* to prevent two processes
				 from trying to write the LCD at the same time */

/* how we handle the interface */
static int lcdgear;

static int lcd_ioctl (struct inode *inode, 
			struct file *file, unsigned int cmd, unsigned long arg)
{
	unsigned int minor;

	minor = MINOR (inode->i_rdev);
	if (MAJOR (inode->i_rdev) != LCD_MAJOR)
		return -EINVAL;
	if (minor >= LCD_MAX)
		return -ENODEV;

	switch (cmd) {
	case LCD_SETCONTROL:	/* switch to control mode */
	        LCD_DATA[minor].mode=MODE_CONTROL;
		break;
	case LCD_SETDATA:       /* switch to data mode */
		LCD_DATA[minor].mode=MODE_DATA;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int lcd_open (struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR (inode->i_rdev);

	if (minor >= LCD_MAX)
		return -ENODEV;	

	if (LCD_DATA[minor].BUSY)
		return -EBUSY;
	LCD_DATA[minor].BUSY = LCD_TRUE;
	MOD_INC_USE_COUNT;
	return 0;
}
#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,31)
static int lcd_release (struct inode *inode, struct file *file)
#else
static void lcd_release (struct inode *inode, struct file *file)
#endif
{
	unsigned int minor = MINOR (inode->i_rdev);
	inode->i_atime = CURRENT_TIME;
	LCD_DATA[minor].BUSY = LCD_FALSE;
	MOD_DEC_USE_COUNT;
#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,31)
	return(0);
#endif
}

/* debounce the status port and return it's value */
static int debounce_status (void)
{
	int dataa, datab;
	
	do { 
		dataa=inb(LCD_S_PORT);
		datab=inb(LCD_S_PORT); 
	} while (dataa != datab);
	
	return(datab);
}

/* 
  
  WARNING if you accidentally unplug the LCD Linux will apear to lock up!
  It's actually stuck waiting for the LCD... This is called a busy wait
  and is rather sloppy, but it does work. This will change in a future
  version and be sceduled. If you want, code in a scheduled wait and mail it,
  and I'll add it in! :-) 
  Alternatly, just short pin 11 to pin 9 and linux should wake up. ;'-)
  
*/

void busy_lcd (void)
{
	int timeout=0;
	if(lcdgear==1){
		udelay(SLEEPER); /* unfortunately needed, and I donno why */
		outb(255-16,LCD_C_PORT); /* read status mode */
        	outb(LCD_BSY,LCD_D_PORT);
		outb(255-LCD_EN-16,LCD_C_PORT); 
		for(;;) { 
		if(!(debounce_status() & 128)) break;
			timeout++;
			if(timeout == 300000) {
				lcdgear=0;
				break;
			}
		}
	
	outb(255-LCD_RW-16,LCD_C_PORT);  /* release mode */
	} else {
	udelay(SLEEPER);
	}
}

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,60)
static ssize_t lcd_write ( struct file *file, const char *buf, 
			   size_t count, loff_t *ppos)
#elif LINUX_VERSION_CODE >= KERNEL_VER(2,1,0)
static long lcd_write (struct inode *inode, 
			struct file *file, const char *buf, 
			unsigned long count)
#else
static int lcd_write (struct inode *inode, 
			struct file *file, const char *buf, 
			int count)
#endif
{
  int cnt, sz;
  struct LCD_DATA_TYPE tempdata;

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,60)
	struct inode *inode = file->f_dentry->d_inode;
#endif

  unsigned int minor;

  if (count != LCD_DATASIZE)
    return -EOVERFLOW;

#if LINUX_VERSION_CODE > KERNEL_VER(2,1,3)
  /* there should be a check here... oh well, fuck it :-) */
  copy_from_user(&tempdata, buf, LCD_DATASIZE);
#else
  verify_area (VERIFY_WRITE, (void *) buf, sizeof (struct LCD_DATA_TYPE));
  memcpy_fromfs (&tempdata, buf, LCD_DATASIZE);
#endif
  
  minor = MINOR (inode->i_rdev);


  cli ();
  if (!LCD_WRITE_SEMAPHORE) {
    LCD_WRITE_SEMAPHORE++;
  }
  sti ();
  
  cnt=0;
  sz=strlen (tempdata.text);
  
  if (LCD_DATA[minor].mode == MODE_CONTROL) {
	CONTROLWRITE(tempdata.text,sz,cnt)
  } else {
	DATAWRITE(tempdata.text,sz,cnt)
  }
  
  LCD_WRITE_SEMAPHORE = 0;	/*allow other reads to progress*/
  
  return LCD_DATASIZE;
}


static struct file_operations lcd_fops = {
	NULL,			/* lcd_lseek*/
	NULL,		        /* lcd_read */
	lcd_write,		/* lcd_write*/
	NULL,			/* lcd_readaddr*/
	NULL,			/* lcd_select */
	lcd_ioctl,		/* lcd_ioctl*/
	NULL,			/* lcd_mmap */
	lcd_open,		/* lcd_open*/
#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,118)
	NULL,			/* lcd_flush */
#endif
	lcd_release,		/* lcd_release*/
	NULL			/* lcd_sync */
};

#ifdef __cplusplus
extern "C" {
#endif

int init_module (void) {
	unsigned char lcd_num;
	unsigned char logon[81];
	int cnt,sz;
	unsigned char *modes[3]={"SLOW","FAST","OLD"};
#ifdef OBSOLETE
	lcdgear=2;
#else
	lcdgear=1;
#endif
	printk ("lcd_init: Hitachi HD44780 LCD driver enabled.\n");
	if (register_chrdev (LCD_MAJOR, "LCD", &lcd_fops))
		  printk ("Unable to get major=%d for LCD\n",
					LCD_MAJOR);
	for (lcd_num = 0; lcd_num < LCD_MAX; lcd_num++)
		  LCD_DATA[lcd_num].BUSY = LCD_FALSE;
	  LCD_WRITE_SEMAPHORE = 0;

	/* Preinitialize the display, show mode, say hello! */
	sprintf(logon,LCD_INITSTRING);
	sz=strlen (logon);     
	cnt=0;	  
	CONTROLWRITE(logon,sz,cnt)	
	
	/* delay a bit */
	udelay(100000);
	sprintf(logon,LCD_CLEARSTRING);
	sz=strlen (logon);     
	cnt=0;
	CONTROLWRITE(logon,sz,cnt)	
	
	/* delay a bit */
	udelay(100000);
	sprintf(logon,"\200"); /* start at 0,0 on the display */
	sz=strlen (logon);     
	cnt=0;	  
	CONTROLWRITE(logon,sz,cnt)	

	/* delay a bit */
	udelay(100000);

	sprintf(logon,"Mode: %u, %s",lcdgear,modes[lcdgear]);
	sz=strlen (logon);
	cnt=0;	  

	DATAWRITE(logon,sz,cnt)	

	/* delay a bit */
	udelay(100000);
	sprintf(logon,"\300"); /* start at 0,1 on the display */
	sz=strlen (logon);
	cnt=0;	  
	CONTROLWRITE(logon,sz,cnt)	

	/* delay a bit */
	udelay(100000);
	sprintf(logon,"Driver Ready.");
	sz=strlen (logon);
	cnt=0;	  
	DATAWRITE(logon,sz,cnt)	

	/* delay a bit */
	udelay(100000);


	  return 0;
}

void cleanup_module (void)
{
	if (MOD_IN_USE) {
/*
We should kill any program using this driver, beacuse
it would make the driver AND the calling program exit cleanly.
*/
		printk ("LCD: device busy, remove delayed.\n");
	}

	if (unregister_chrdev (LCD_MAJOR, "LCD") != 0) {
		printk ("cleanup_module failed.\n");
	}
	else {
		printk ("cleanup_module succeeded.\n");
	}
}



#ifdef __cplusplus
}
#endif

