/*

lcdprocinfo: process status functions
Copyleft 1997 Andrew J. Kroll (ag784@freenet.buffalo.edu)

*/

#include "lcd.h"

#include "proc/ps.h"
#include "proc/readproc.h"

#include "top.h"

static proc_t **p_table=NULL;
static int sleeping=0, stopped=0, zombie=0, running=0, 
				total=0, disksleep=0, paging=0; 


/* how do we read the process table? top style (VOMIT), or the kernel hack? */
int proc_get_method(void)
{
	int junk=-1;
	char stuff[40];
	FILE *here;
		
	here=fopen("/proc/realtime/processes/prstcnts", "r");
	if (here) {
		fclose(here);
		printf("Using RT module mode!\n");
		return(2); /* use kernel module! YAY! */
	}

	fclose(here); /* sanity check */

	here=fopen("/proc/stat", "r");
  	do {
		fgets(stuff, 40, here);
	} while (!feof(here) && strncmp(stuff, "prstcnts", 10));
	sscanf(stuff, "prstcnts %d", &junk );
	fclose(here);
	if (junk != -1) { 
		printf("Yuck! Old kernel hack mode!\n");
		return(1); /* use kernel hack! */
		
	}
	printf("Yuck! Top style mode!\n");
	return(0);
}

/*#####################################################################
 *#######   A readproctable function that uses already allocated  #####
 *#######   table entries.                                        #####
 *#####################################################################
 */
#define Do(x) (flags & PROC_ ## x)

float get_elapsed_time(void)
{
	struct timeval time;
	static struct timeval oldtime;
	struct timezone timez;
	float elapsed_time;

	gettimeofday(&time, &timez);
	elapsed_time = (time.tv_sec - oldtime.tv_sec)+
	(float) (time.tv_usec - oldtime.tv_usec) / 1000000.0;
	oldtime.tv_sec = time.tv_sec;
	oldtime.tv_usec = time.tv_usec;
	return (elapsed_time);
}

proc_t** readproctab2(int flags, proc_t** tab, ...)
{
	PROCTAB* PT = NULL;
	static proc_t *buff;
	int n = 0;
	static int len = 0;
	va_list ap;

	va_start(ap, tab);      /* pass through args to openproc */
	if (Do(UID))
		PT = openproc(flags, va_arg(ap, uid_t*), va_arg(ap, int));
	else if (Do(PID) || Do(TTY) || Do(STAT))
		PT = openproc(flags, va_arg(ap, void*)); /* assume ptr sizes same */
	else
		PT = openproc(flags);
	va_end(ap);
	buff = (proc_t *) 1;
	while (n<len && buff) {     /* read table: (i) already allocated chunks */
		if (tab[n]->cmdline) {
			free((void*)*tab[n]->cmdline);
			tab[n]->cmdline = NULL;
		}
		buff = readproc(PT, tab[n]);
		n++;
	}
	if (buff) {
		do {               /* (ii) not yet allocated chunks */
			tab = realloc(tab, (n+1)*sizeof(proc_t*));/* realloc as we go, using */
			buff = readproc(PT, NULL);        /* final null to terminate */
			if(buff) tab[n]=buff;
			len++;
			n++;
		} while (buff);           /* stop when NULL reached */
		tab[n-1] = xcalloc(NULL, sizeof (proc_t));
		tab[n-1]->pid=-1;        /* Mark end of Table */
	} else {
		if (n == len) {
			tab = realloc(tab, (n+1)*sizeof(proc_t*));
			tab[n] = xcalloc(NULL, sizeof (proc_t));
			len++;
		}
		tab[n]->pid=-1;    /* Use this instead of NULL when not at the end of */
	}                   /* the allocated space */
	closeproc(PT);
	return tab;
}


/*
 * Calculates the number of processes running (using old icky top code)
 */
void do_stats (proc_t** p, float elapsed_time, int pass,int fd, int *sleeping, int *stopped , int *zombie , int *running)
{
	proc_t *this;
	int index, n = 0 ,i = 0;
	static int prev_count = 0;
	static struct save_hist save_history[NR_TASKS];
	struct save_hist New_save_hist[NR_TASKS];

	*sleeping = *stopped = *zombie = *running = n = 0;

	/*
	 * Make a pass through the data to get stats.
	 */
	index = 0;
	while (p[n]->pid != -1) {
		this = p[n];
		switch (this->state) {
			case 'S':
			case 'D':
				(*sleeping)++;
				break;
			case 'T':
				(*stopped)++;
				break;
			case 'Z':
				(*zombie)++;
				break;
			case 'R':
				(*running)++;
				break;
			default:
				/* Don't know how to handle this one. */
				break;
		}
		index++;
		n++;
	}

	/*
	 * Save this frame's information.
	 */
	for (i = 0; i < n; i++) {
		/* copy the relevant info for the next pass */
		save_history[i].pid = New_save_hist[i].pid;
	}

	prev_count = n;
	qsort(p, n, sizeof(proc_t*), (void*)mult_lvl_cmp);
}


/*
 * Calculates number of processes running (using phil's cool kernel mod!) :)
 */
void do_stats_new(int *nrunning, int *nsleeping, int *ndisk_sleep, 
					int *nzombie, int *nstopped, 
					int *npaging)
{
	FILE *foo;
	char fuck_me[40];
	
	foo = fopen("/proc/stat", "r");
	do { fgets(fuck_me, 40, foo); 
	} while (!feof(foo) && strncmp(fuck_me, "prstcnts", 10));
	sscanf(fuck_me, "prstcnts %d %d %d %d %d %d", nrunning, nsleeping,
				ndisk_sleep, nzombie, nstopped, npaging);
	fclose(foo);
} /* do_proc_stats --PS */

/* get stats via my module */
void do_stats_module(int *nrunning, int *nsleeping, int *ndisk_sleep, 
					int *nzombie, int *nstopped, 
					int *npaging)
{
	FILE *foo;
	char fuck_me[40];
	
	foo = fopen("/proc/realtime/processes/prstcnts", "r");
	do { fgets(fuck_me, 40, foo); 
	} while (!feof(foo) && strncmp(fuck_me, "prstcnts", 10));
	sscanf(fuck_me, "prstcnts %d %d %d %d %d %d", nrunning, nsleeping,
				ndisk_sleep, nzombie, nstopped, npaging);
	fclose(foo);
}

/* update the process tables */
void update_proc (int fd, int method)
{
	if (method==2) {
		do_stats_module(&running, &sleeping, &disksleep,
					&zombie, &stopped, &paging);
	} else if (method==1) { 
		do_stats_new(&running, &sleeping, &disksleep,
					&zombie, &stopped, &paging); 
	} else if (method==0) {
		p_table=readproctab2(proc_flags, p_table, NULL);
		elapsed_time = get_elapsed_time();
		do_stats(p_table, elapsed_time, 0, fd, &sleeping,
						&stopped, &zombie, &running);
	}	
	total=sleeping + stopped + zombie + running;
}

void draw_proc (int fd,int pass)
{
	char ln[41];
	int spot;
	/* hmmm we'll need a formatted output for these, but for now
	lets stick to the basics... and get it working first. */

	spot=(te_procinfo[pass][1] * 64) + te_procinfo[pass][0];
	lcd_pos(fd, (spot & 0xff));

	switch(te_procinfo[pass][3]) {

		case 1:
			sprintf (ln, "S%.3i", sleeping);
			break;

		case 2:
			sprintf (ln, "H%.3i", stopped);
			break;

		case 3:
			sprintf (ln, "Z%.3i", zombie);
			break;

		case 4:
			sprintf (ln, "R%.3i", running);
			break;

		case 5:
			sprintf (ln, "T%.3i", total);
			break;

		case 6:
			sprintf (ln, "D%.3i", disksleep);
			break;

		case 7:
			sprintf (ln, "P%.3i", paging);
			break;

		default:
			sprintf (ln, "?XXX");
			break;

	}
	lcd_write (fd, ln);
		
}


