/*

Copyleft 1997 Andrew J. Kroll (ag784@freenet.buffalo.edu)

06/02/98	Andrew J. Kroll 0.01   first version
*/

#define DRIVER
#include <linux/fdreg.h>

#include <linux/mm.h>
#include <linux/list.h>
#include <linux/sched.h>

#include <linux/slab.h>
#include <linux/init.h>

#include "lcd.h"
#if LINUX_VERSION_CODE >= KERNEL_VER(2,2,0)
#include <asm/current.h>
#endif

#include <linux/proc_fs.h>
#include <linux/linkage.h>
#include <linux/signal.h>

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,16)
#include <linux/init.h>
#else
#define __initdata
#define __initlocaldata
#define __initfunc(__arg) __arg
#define __init
#endif


#if LINUX_VERSION_CODE < KERNEL_VER(2,2,11) /* bail out */

#else

#if LINUX_VERSION_CODE < KERNEL_VER(2,1,28)

#include <asm/segment.h> /* for memcpy_tofs() */

struct proc_dir_entry proc_rt;

static struct file_operations prst_proc_fops = {
	NULL,			/* prst_lseek*/
	prst_read,		/* prst_read */
	NULL,			/* prst_write*/
	NULL,			/* prst_readaddr*/
	NULL,			/* prst_select */
	NULL,			/* prst_ioctl*/
	NULL,			/* prst_mmap */
	NULL,			/* prst_open*/
#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,118)
	NULL,			/* prst_flush */
#endif
	NULL,			/* prst_release*/
	NULL			/* prst_sync */
};

static struct inode_operations prst_proc_inode_operations =
{
	&prst_proc_fops,
	NULL,		/* create */
	NULL,		/* lookup */
	NULL,		/* link */
	NULL,		/* unlink */
	NULL,		/* symlink */
	NULL,		/* mkdir */
	NULL,		/* rmdir */
	NULL,		/* mknod */
	NULL,		/* rename */
	NULL,		/* readlink */
	NULL,		/* readpage */
	NULL,		/* writepage */
	NULL,		/* bmap */
	NULL,		/* truncate */
	NULL		/* permission */
}


/*  Read prst proc directory entry.
 */

#define PROC_BLOCK_SIZE PAGE_SIZE

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,0)  
static long prst_proc_read(struct inode * inode, struct file * file,
				char * buf, unsigned long nbytes)
#else
static int prst_proc_read(struct inode * inode, struct file * file,
				char * buf, int nbytes)
#endif
{
	char    *page;
	int     retval=0;
	int     eof=0;
	int     n, count;
	char    *start;
	struct proc_dir_entry * dp;
	
	if (nbytes < 0)    
		return -EINVAL;
	dp = (struct proc_dir_entry *) inode->u.generic_ip;
	if (!(page = (char*) __get_free_page(GFP_KERNEL)))
		return -ENOMEM;
	
	while ((nbytes > 0) && !eof)
	{                  
		count = PROC_BLOCK_SIZE <= nbytes ? PROC_BLOCK_SIZE : nbytes;
		
		start = NULL;
		if (dp->get_info) {
			/* 
			 * Handle backwards compatibility with the old net
			 * routines.
			 * 
			 * XXX What gives with the file->f_flags & O_ACCMODE
			 * test?  Seems stupid to me....
			 */
			n = dp->get_info(page, &start, file->f_pos, count,
				(file->f_flags & O_ACCMODE) == O_RDWR);
			if (n < count)
				eof = 1;
		} else     
			break;
		
		if (!start) {
			/* 
			 * For proc files that are less than 4k
			 */
			start = page + file->f_pos;
			n -= file->f_pos;
			if (n <= 0)
				break;
			if (n > count)
				n = count;
		}
		if (n == 0)
			break;  /* End of file */
		if (n < 0) {
			if (retval == 0)
				retval = n;
			break;
		}
#if LINUX_VERSION_CODE > KERNEL_VER(2,1,3)
		copy_to_user(buf, start, n);
#else                      
		memcpy_tofs(buf, start, n);
#endif                     
		file->f_pos += n;       /* Move down the file */
		nbytes -= n;
		buf += n;  
		retval += n;
	}
	free_page((unsigned long) page);
	return retval;     
}

#else /* LINUX_VERSION_CODE < KERNEL_VER(2,1,28) */

#define PRST_PROC_REGISTER(parent, child) proc_register(parent, child)

#if LINUX_VERSION_CODE >= KERNEL_VER(2,1,58)
#define __NO_VERSION__     
#include <linux/module.h>  
/*                         
 * This is called as the fill_inode function when an inode
 * is going into (fill = 1) or out of service (fill = 0).
 * We use it here to manage the module use counts.
 *                         
 * Note: only the top-level directory needs to do this; if
 * a lower level is referenced, the parent will be as well.
 */                        
static void proc_prst_fill_inode(struct inode *inode, int fill)
{                          
	if (fill) {        
		MOD_INC_USE_COUNT;
	} else {
		MOD_DEC_USE_COUNT;
	}
}                          
#else                      
#define proc_prst_fill_inode NULL
#endif                     
/*                         
 * Proc filesystem directory entries.
 */                        

static struct proc_dir_entry proc_rt =
{                          
	0,				/* low_ino */
	sizeof("realtime") - 1,		/* namelen */
	"realtime",			/* name */
	S_IFDIR | S_IRUGO | S_IXUGO,	/* mode */
	2,				/* nlink */
	0,				/* uid */
	0,				/* gid */
	0,				/* size */
	NULL,				/* ops */
	NULL,				/* get_info */
	proc_prst_fill_inode,		/* fill_inode */
	NULL,				/* next */
	NULL,				/* parent */
	NULL,				/* subdir */
	NULL,				/* data */
	NULL,				/* read_proc  */
	NULL				/* write_proc */
};                         



#define RT_PROC_ENTRY(dir, name)                     \
static struct proc_dir_entry proc_rt_ ##dir ## _ ##name =       \
{                                                        \
        0,                           /* low_ino */	\
        sizeof(#name) - 1,           /* namelen */	\
        #name,                       /* name */		\
        S_IFDIR | S_IRUGO | S_IXUGO, /* mode */		\
        2,                           /* nlink */	\
        0,                           /* uid */		\
        0,                           /* gid */		\
        0,                           /* size */		\
        NULL,                        /* ops */		\
        NULL,                        /* get_info */	\
        NULL,                        /* fill_inode */	\
        NULL,                        /* next */		\
        NULL,                        /* parent */	\
        NULL,                        /* subdir */	\
        NULL,                        /* data */		\
        NULL,                        /* read_proc  */	\
        NULL                         /* write_proc */	\
};

#define PRST_PROC_ENTRY(dir, name)			\
static struct proc_dir_entry proc_rt_ ##dir ## _ ##name =	\
{							 \
        0,                         /* low_ino    */	\
        sizeof(#name)-1,           /* namelen    */	\
        #name,                     /* name       */	\
        S_IFREG | S_IRUGO,         /* mode       */	\
        1,                         /* nlink      */	\
        0,                         /* uid        */	\
        0,                         /* gid        */	\
        0,                         /* size       */	\
        NULL,                      /* ops        */	\
        NULL,                      /* get_info   */	\
        NULL,                      /* fill_inode */	\
        NULL,                      /* next       */	\
        NULL,                      /* parent     */	\
        NULL,                      /* subdir     */	\
        NULL,                      /* data       */	\
        pr_read_proc_##dir ## _ ##name, /* read_proc  */	\
        NULL                       /* write_proc */	\
}

#define PRST_PROC_DIR_ENTRY(name) \
static struct proc_dir_entry proc_rt_##name = \
{ \
	0, 				/* low_ino */ \
	sizeof(#name)-1,		/* namelen */ \
	#name,				/* name */ \
	S_IFDIR | S_IRUGO | S_IXUGO,	/* mode */ \
	2,				/* nlink */ \
	0,				/* uid */ \
	0,				/* gid */ \
	0,				/* size */ \
	NULL,				/* ops */ \
	NULL,				/* get_info */ \
	NULL,				/* fill_inode */ \
	NULL,				/* next */ \
	NULL,				/* parent */ \
	NULL,				/* subdir */ \
	NULL,				/* data */ \
};
	
#endif

int prst_read_proc_prstcnts(int sel, char *page, char **start, off_t off,
			      int count, int *eof, void *data)
{
	unsigned sleeping=0, zombie=0, stopped=0, running=0, disk_sleep=0, paging=0;
	char *ptr = page;
	size_t len;
	struct task_struct *p;
	
	*start = 0;
	read_lock(&tasklist_lock);
	for_each_task(p) {
		switch(p->state) {
 			case 0:
	 			running++;
 				break;
	 		case 1:
	 			sleeping++;
	 			break;
	 		case 2:
	 			disk_sleep++;
	 			break;
	 		case 3:
	 			zombie++;
	 			break;
	 		case 4:
	 			stopped++;
	 			break;
	 		case 5:
	 			paging++;
	 			break;
	 		default:
	 			/* eh!? */
	 			break;
		}
	}
	read_unlock(&tasklist_lock);

	len = sprintf(ptr,
		     "prstcnts %u %u %u %u %u %u\n",
		     running, sleeping, disk_sleep, zombie, stopped, paging
		     );

	if (off + count >= len) {
		*eof = 1;
	} else {
		*eof = 0;
	} 
	return len;
}


#define PRST_READ_PROC_PRSTCNTS(sel) \
static int pr_read_proc_##sel ## _prstcnts(char *page, char **start, off_t off, \
	int count, int *eof, void *data) \
{ \
	return prst_read_proc_prstcnts(##sel, page, start, off, count, eof, data); \
}

static int processes;

PRST_READ_PROC_PRSTCNTS(processes);
PRST_PROC_DIR_ENTRY(processes);
PRST_PROC_ENTRY(processes, prstcnts);



int prst_proc_init(void)
{
#if LINUX_VERSION_CODE < KERNEL_VER(2,1,28)
        proc_prst.ops =
		proc_rt_processes.ops;
#endif

	if (PRST_PROC_REGISTER(&proc_root, &proc_rt) < 0) {
		return -EIO;
	}
	if (PRST_PROC_REGISTER(&proc_rt, &proc_rt_processes) >= 0) {
		PRST_PROC_REGISTER(&proc_rt_processes, &proc_rt_processes_prstcnts);
	}
	return 0;
}

void prst_proc_destroy(void)
{
        if (proc_rt.low_ino != 0) {
		if (proc_rt_processes.low_ino != 0) {
			proc_unregister(&proc_rt_processes, proc_rt_processes_prstcnts.low_ino);
		}
	}
	proc_unregister(&proc_root, proc_rt.low_ino);
}

int init_module (void) {
	  return (prst_proc_init());
}

void cleanup_module (void)
{
	prst_proc_destroy();
}

#endif
